import os

from hotsos.core.log import log
from hotsos.core.ycheck.engine.properties.inputdef import YPropertyInputBase
from hotsos.core.ycheck.engine.properties.requires import (
    intercept_exception,
    YRequirementTypeBase,
)


class YRequirementTypePath(YPropertyInputBase, YRequirementTypeBase):
    """ Provides logic to perform checks on filesystem paths. """

    @classmethod
    def _override_keys(cls):
        # We can't use 'input' since that property is already used.
        return ['path']

    @property
    def options(self):
        """
        Override this since we never want to have all-logs applied since
        it is not relevant in checking if the path exists.
        """
        _options = super().options
        _options['disable-all-logs'] = True
        return _options

    @property
    @intercept_exception
    def _result(self):
        _result = True
        not_found = None
        for path in self.paths:
            if not os.path.exists(path):
                not_found = path
                _result = False
                break

        log.debug('requirement check: path %s (result=%s)', not_found,
                  _result)
        self.cache.set('path_not_found', not_found)
        self.cache.set('paths', self.paths)
        return _result



import datetime

import yaml
from hotsos.core.analytics import LogEventStats, SearchResultIndices
from hotsos.core.config import HotSOSConfig
from hotsos.core.host_helpers import CLIHelper
from hotsos.core.issues import (
    IssueContext,
    IssuesManager,
    NeutronL3HAWarning,
    OpenstackWarning
)
from hotsos.core.log import log
from hotsos.core.plugins.openstack.common import (
    OpenstackChecksBase,
    OpenstackEventChecksBase,
)
from hotsos.core.plugins.openstack.neutron import NeutronHAInfo
from hotsos.core.plugintools import summary_entry_offset as idx
from hotsos.core.search import (
    FileSearcher,
    SearchConstraintSearchSince,
)
from hotsos.core import utils
from hotsos.core.utils import sorted_dict
from hotsos.core.ycheck.events import CallbackHelper
from hotsos.core.ycheck.engine.properties.search import CommonTimestampMatcher

EVENTCALLBACKS = CallbackHelper()
VRRP_TRANSITION_WARN_THRESHOLD = 8


class ApacheEventChecks(OpenstackEventChecksBase):

    def __init__(self, *args, **kwargs):
        super().__init__(EVENTCALLBACKS, *args, **kwargs)

    @property
    def root_group_name(self):
        return 'apache'

    @EVENTCALLBACKS.callback(event_group='apache')
    def connection_refused(self, event):
        ports_max = {}
        context = {}

        results = []
        for result in event.results:
            # save some context info
            path = self.searcher.resolve_source_id(result.source_id)
            if path in context:
                context[path].append(result.linenumber)
            else:
                context[path] = [result.linenumber]

            month = datetime.datetime.strptime(result.get(1), '%b').month
            day = result.get(2)
            year = result.get(3)
            addr = result.get(4)
            port = result.get(5)
            results.append({'date': "{}-{}-{}".format(year, month, day),
                            'key': "{}:{}".format(addr, port)})

        conns_refused = self.categorise_events(event, results=results)
        for addrs in conns_refused.values():
            for addr, count in addrs.items():
                port = addr.partition(':')[2]
                # allow a small number of connection refused errors on a given
                # day
                if count < 5:
                    continue

                if port not in ports_max:
                    ports_max[port] = count
                else:
                    ports_max[port] = max(count, ports_max[port])

        if ports_max:
            msg = ('apache is reporting connection refused errors for the '
                   'following ports which could mean some services are not '
                   'working properly - {} - please check.'.
                   format(','.join(ports_max.keys())))
            IssuesManager().add(OpenstackWarning(msg), IssueContext(**context))

        return sorted_dict(conns_refused)

    def __summary_apache(self):
        return self.final_event_results


class APIEvents(OpenstackEventChecksBase):

    def __init__(self, *args, **kwargs):
        super().__init__(EVENTCALLBACKS, *args, **kwargs)

    @property
    def root_group_name(self):
        return 'http-requests'

    @EVENTCALLBACKS.callback(event_group='http-requests',
                             event_names=['neutron'])
    def http_requests(self, event):
        results = [{'date': r.get(1),
                    'key': r.get(2)} for r in event.results]
        ret = self.categorise_events(event, results=results)
        if ret:
            return ret

    def __summary_http_requests(self):
        out = self.final_event_results
        if out:
            return out


class NeutronAgentEventChecks(OpenstackEventChecksBase):
    """
    Loads events we want to check from definitions yaml and executes them. The
    results are sorted by date and the "top 5" are presented along with stats
    on the full set of samples.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(EVENTCALLBACKS, *args, **kwargs)

    @property
    def root_group_name(self):
        return 'neutron.agents'

    def _get_event_stats(self, results, tag_prefix, custom_idxs=None):
        stats = LogEventStats(results, tag_prefix, custom_idxs=custom_idxs)
        stats.run()
        top5 = stats.get_top_n_events_sorted(5)
        if not top5:
            return

        return {"top": top5,
                "stats": stats.get_event_stats()}

    @EVENTCALLBACKS.callback(event_group='neutron.agents')
    def router_updates(self, event):
        agent = event.section
        sri = SearchResultIndices(event_id_idx=4,
                                  metadata_idx=3,
                                  metadata_key='router')
        ret = self._get_event_stats(event.results, event.search_tag,
                                    custom_idxs=sri)
        if ret:
            return {event.name: ret}, agent

    @EVENTCALLBACKS.callback(event_group='neutron.agents',
                             event_names=['rpc-loop', 'router-spawn-events'])
    def process_events(self, event):
        agent = event.section
        ret = self._get_event_stats(event.results, event.search_tag)
        if ret:
            return {event.name: ret}, agent

    def __summary_neutron_l3_agent(self):
        out = self.final_event_results or {}
        return out.get('neutron-l3-agent')

    def __summary_neutron_ovs_agent(self):
        out = self.final_event_results or {}
        return out.get('neutron-ovs-agent')


class OctaviaAgentEventChecks(OpenstackEventChecksBase):

    def __init__(self, *args, **kwargs):
        super().__init__(EVENTCALLBACKS, *args, **kwargs)

    @property
    def root_group_name(self):
        return 'octavia'

    @EVENTCALLBACKS.callback(event_group='octavia',
                             event_names=['lb-failover-auto',
                                          'lb-failover-manual'])
    def lb_failovers(self, event):
        results = []
        for e in event.results:
            payload = yaml.safe_load(e.get(3))
            lb_id = payload.get('load_balancer_id')
            if lb_id is None:
                continue

            results.append({'date': e.get(1), 'time': e.get(2), 'key': lb_id})

        ret = self.categorise_events(event, results=results, key_by_date=False)
        if ret:
            failover_type = event.name.rpartition('-')[2]
            return {failover_type: ret}, 'lb-failovers'

    @EVENTCALLBACKS.callback(event_group='octavia')
    def amp_missed_heartbeats(self, event):
        missed_heartbeats = self.categorise_events(event, key_by_date=False)
        if not missed_heartbeats:
            return

        # sort each amp by occurences
        for ts_date, amps in missed_heartbeats.items():
            missed_heartbeats[ts_date] = utils.sorted_dict(amps,
                                                           key=lambda e: e[1],
                                                           reverse=True)

        # then sort by date
        return utils.sorted_dict(missed_heartbeats)

    def __summary_octavia(self):
        return self.final_event_results


class NovaComputeEventChecks(OpenstackEventChecksBase):

    def __init__(self, *args, **kwargs):
        super().__init__(EVENTCALLBACKS, *args, **kwargs)

    @property
    def root_group_name(self):
        return 'nova.nova-compute'

    @EVENTCALLBACKS.callback(event_group='nova.nova-compute')
    def pci_dev_not_found(self, event):
        ret = self.categorise_events(event)
        if ret:
            return ret, 'PciDeviceNotFoundById'

    def __summary_nova(self):
        return self.final_event_results


class AgentApparmorChecks(OpenstackEventChecksBase):

    def __init__(self, *args, **kwargs):
        super().__init__(EVENTCALLBACKS, *args, **kwargs)

    @property
    def root_group_name(self):
        return 'apparmor'

    @EVENTCALLBACKS.callback(event_group='apparmor',
                             event_names=['nova', 'neutron'])
    def openstack_apparmor(self, event):
        results = [{'date': "{} {}".format(r.get(1), r.get(2)),
                    'time': r.get(3),
                    'key': r.get(4)} for r in event.results]
        ret = self.categorise_events(event, results=results,
                                     key_by_date=False)
        if ret:
            # event.name must be the service name, event.section is the aa
            # action.
            return {event.name: ret}, event.section

    def __summary_apparmor(self):
        return self.final_event_results


class NeutronL3HAEventChecks(OpenstackEventChecksBase):

    def __init__(self, *args, **kwargs):
        super().__init__(EVENTCALLBACKS, *args, **kwargs)
        self.cli = CLIHelper()
        self.ha_info = NeutronHAInfo()

    @property
    def root_group_name(self):
        return 'neutron.ml2-routers'

    def check_vrrp_transitions(self, transitions):
        # there will likely be a large number of transitions if we look across
        # all time so dont run this check.
        if HotSOSConfig.use_all_logs:
            return

        max_transitions = 0
        warn_count = 0
        threshold = VRRP_TRANSITION_WARN_THRESHOLD
        for router in transitions:
            r = transitions[router]
            _transitions = sum(t for d, t in r.items())
            if _transitions > threshold:
                max_transitions = max(_transitions, max_transitions)
                warn_count += 1

        if warn_count:
            msg = ("{} router(s) have had more than {} vrrp transitions "
                   "(max={}) in the last 24 hours.".format(warn_count,
                                                           threshold,
                                                           max_transitions))
            IssuesManager().add(NeutronL3HAWarning(msg))

    def journalctl_args(self):
        """ Args callback for event cli command """
        args = []
        kwargs = {'unit': 'neutron-l3-agent'}
        if not HotSOSConfig.use_all_logs:
            kwargs['date'] = self.cli.date(format="--iso-8601")

        return args, kwargs

    @EVENTCALLBACKS.callback(event_group='neutron.ml2-routers')
    def vrrp_transitions(self, event):
        results = []
        for r in event.results:
            router = self.ha_info.find_router_with_vr_id(r.get(3))
            if not router:
                log.debug("could not find router with vr_id %s", r.get(3))
                continue

            results.append({'date': r.get(1), 'time': r.get(2),
                            'key': router.uuid})

        transitions = self.categorise_events(event, results=results,
                                             key_by_date=False)
        if transitions:
            # run checks
            self.check_vrrp_transitions(transitions)
            # add info to summary
            return {'transitions': transitions}, 'keepalived'

    def __summary_neutron_l3ha(self):
        return self.final_event_results


class AgentEventChecks(OpenstackChecksBase):

    def _run_checks(self, checks):
        # Only run if we think Openstack is installed.
        if not self.openstack_installed:
            return

        searcher = FileSearcher(constraint=SearchConstraintSearchSince(
                                        ts_matcher_cls=CommonTimestampMatcher))
        check_objs = [c(searcher=searcher) for c in checks]
        for check in check_objs:
            check.load()

        results = searcher.run()
        _final_results = {}
        for check in check_objs:
            check.run(results)
            check_results = check.raw_output
            if check_results:
                _final_results.update(check_results)

        return _final_results

    @idx(1)
    def __summary_api_info(self):
        checks = [APIEvents]
        results = self._run_checks(checks)
        if results:
            return results

    @idx(2)
    def __summary_agent_checks(self):
        checks = [AgentApparmorChecks,
                  ApacheEventChecks,
                  NeutronL3HAEventChecks,
                  NeutronAgentEventChecks,
                  NovaComputeEventChecks,
                  OctaviaAgentEventChecks]
        results = self._run_checks(checks)
        if results:
            return results



import os
from datetime import datetime

from cryptography.hazmat.backends import default_backend
from cryptography import x509
from hotsos.core.config import HotSOSConfig
from hotsos.core.factory import FactoryBase
from hotsos.core.host_helpers.cli import CLIHelper
from hotsos.core.log import log


class SSLCertificate(object):

    def __init__(self, certificate_path):
        self.path = os.path.join(HotSOSConfig.data_root, certificate_path)
        try:
            with open(self.path, "rb") as fd:
                self.certificate = fd.read()
        except OSError as e:
            log.warning("Unable to read SSL certificate file %s: %s",
                        self.path, e)
            raise

    @property
    def expiry_date(self):
        """
        Return datetime() of when the certificate expires
        """

        cert = x509.load_pem_x509_certificate(self.certificate,
                                              default_backend())
        return cert.not_valid_after

    @property
    def days_to_expire(self):
        """
        Return int(days) remaining until the certificate expires
        """
        fmt = '%Y-%m-%d %H:%M:%S'
        today = datetime.strptime(CLIHelper().date(format='+' + fmt),
                                  fmt)
        days = self.expiry_date - today
        return int(days.days)


class SSLCertificatesHelper(object):

    def __init__(self, certificate, expire_days):
        """
        @param certificate: SSLCertificate object.
        @param expire_days: int(expire_days) used to check the
                            days that remain until certificate expiration.
        """
        self.certificate = certificate
        self.expire_days = expire_days

    @property
    def certificate_expires_soon(self):
        """
        @return: True if certificate expires in less than int(self.expire_days)
        """
        return self.certificate.days_to_expire <= self.expire_days


class SSLCertificatesFactory(FactoryBase):
    """
    Factory to dynamically create SSLCertificate objects for given paths.

    SSLCertificate objects are returned when a getattr() is done on this object
    using cert path as the attr name.
    """

    def __getattr__(self, path):
        log.debug("creating SSLCertificate object for %s", path)
        return SSLCertificate(path)



from datetime import datetime

from hotsos.core.analytics import LogEventStats
from hotsos.core.plugins.openstack.common import (
    OpenstackChecksBase,
    OpenstackEventChecksBase,
)
from hotsos.core.plugins.openstack.nova import NovaLibvirt
from hotsos.core import utils
from hotsos.core.ycheck.events import CallbackHelper

EVENTCALLBACKS = CallbackHelper()


class OpenstackInstanceChecks(OpenstackChecksBase):

    def __summary_vm_info(self):
        _info = {}

        instances = self.nova.instances.values()
        if instances:
            _info['running'] = {'count': len(instances),
                                'uuids': [i.uuid for i in instances]}

        novalibvirt = NovaLibvirt()
        cpu_models = novalibvirt.cpu_models
        if cpu_models:
            _info["cpu-models"] = cpu_models

        vm_vcpu_info = novalibvirt.vcpu_info
        if vm_vcpu_info:
            _info["vcpu-info"] = vm_vcpu_info

        if _info:
            return _info


class NovaServerMigrationAnalysis(OpenstackEventChecksBase):

    def __init__(self, *args, **kwargs):
        super().__init__(EVENTCALLBACKS, *args, **kwargs)

    @property
    def root_group_name(self):
        return 'nova.migrations'

    def migration_seq_info(self, event, resource_idx, info_idxs,
                           incl_time_in_date=False):
        """
        Process the results of an event that was defined as a sequence.

        Returns a dict keyed by a resource id inside which is a dict of
        information keyed by date. The date can optionally be extended to
        include the time to get greater granularity - requires the associated
        search patterns to include this information.

        @param event: EventCheckResult object
        @param resource_idx: int index of the search result group that is used
                             as the resource identifier.
        @param info_idxs: dict if info types and indexes for arbitrary
                          information to be save against each resource.
        @param incl_time_in_date: include the time in the date string. Defaults
                                  to False.
        """
        info = {}
        for section in event.results:
            for result in section:
                if not result.tag.endswith('-body'):
                    continue

                ts_date = result.get(1)
                if incl_time_in_date:
                    ts_time = result.get(2)
                    ts_date = "{} {}".format(ts_date, ts_time)

                resource = result.get(resource_idx)
                if resource not in info:
                    info[resource] = {}

                if result.section_id not in info[resource]:
                    info[resource][result.section_id] = {}

                if ts_date not in info[resource]:
                    info[resource][result.section_id][ts_date] = {}

                resource_date = info[resource][result.section_id][ts_date]

                for name, idx in info_idxs.items():
                    value = result.get(idx)
                    if value is None:
                        continue

                    if name not in resource_date:
                        resource_date[name] = [value]
                    else:
                        resource_date[name].append(value)

        return info

    @EVENTCALLBACKS.callback(event_group='nova.migrations')
    def src_migration(self, event):
        """
        Source migration is defined as a sequence so that we can capture some
        of in the interim events such as memory and disk progress.
        """
        migration_info = {}

        info_idxs = {'memory': 4, 'disk': 5}
        results = self.migration_seq_info(event, 3, info_idxs,
                                          incl_time_in_date=True)
        for vm_uuid, sections in results.items():
            for section in sections.values():
                samples = {}
                start = None
                end = None
                for date, info in utils.sorted_dict(section).items():
                    if start is None:
                        start = date

                    end = date
                    for rtype, values in info.items():
                        if rtype not in samples:
                            samples[rtype] = []

                        samples[rtype] += [int(i) for i in values]

                _start = datetime.strptime(start, "%Y-%m-%d %H:%M:%S")
                _end = datetime.strptime(end, "%Y-%m-%d %H:%M:%S")
                duration = round(float((_end - _start).total_seconds()), 2)
                info = {'start': start, 'end': end, 'duration': duration}
                instance = self.nova.instances.get(vm_uuid)
                if instance and instance.memory_mbytes is not None:
                    info['resources'] = {'memory_mbytes':
                                         instance.memory_mbytes}

                if samples:
                    # regressions imply that the progress counter had one or
                    # more decreases before increasing again.
                    info['regressions'] = {}
                    for rtype, values in samples.items():
                        if 'iterations' not in info:
                            info['iterations'] = len(values)

                        loops = utils.sample_set_regressions(values)
                        info['regressions'][rtype] = loops

                if vm_uuid in migration_info:
                    migration_info[vm_uuid].append(info)
                else:
                    migration_info[vm_uuid] = [info]

        # section name expected to be live-migration
        return migration_info, event.section

    def migration_stats_info(self, event):
        """
        Process events that have passthrough-results=True such that they can be
        passed directory to analytics.LogEventStats for parsing.
        """
        stats = LogEventStats(event.results, event.name)
        stats.run()
        top5 = stats.get_top_n_events_sorted(5)
        if not top5:
            return

        results = {"top": top5}
        # There can be a very large number of incomplete migrations so need to
        # find a useful way to represent this
        # if stats.data.incomplete_events:
        #     results['incomplete-migrations'] = stats.data.incomplete_events

        return results

    @EVENTCALLBACKS.callback(event_group='nova.migrations')
    def src_post_live_migration(self, event):
        # section name expected to be live-migration
        return self.migration_stats_info(event), event.section

    @EVENTCALLBACKS.callback(event_group='nova.migrations')
    def dst_pre_live_migration(self, event):
        # section name expected to be live-migration
        return self.migration_stats_info(event), event.section

    def __summary_nova_migrations(self):
        return self.load_and_run()



#!/usr/bin/python3
import logging
import tempfile

from hotsos.core.config import HotSOSConfig

log = logging.getLogger('hotsos')


def setup_logging(level=logging.DEBUG):
    fmt = ("%(asctime)s.%(msecs)03d %(process)d %(levelname)s %(name)s [-] "
           "%(message)s")
    log.name = 'hotsos.plugin.{}'.format(HotSOSConfig.plugin_name)
    log.setLevel(level)

    if not log.hasHandlers():
        if HotSOSConfig.debug_mode:
            handler = logging.StreamHandler()
        else:
            handler = logging.FileHandler(tempfile.mktemp(suffix='hotsos.log'))

        handler.setFormatter(logging.Formatter(fmt))
        log.addHandler(handler)

        # Set logging level for dependencies
        logging.getLogger('searchkit').setLevel(level=level)
        logging.getLogger('searchkit').addHandler(handler)
        # NOTE: dont set log level here as it is controlled using env var
        logging.getLogger('propertree').addHandler(handler)



import re
from datetime import timedelta
from functools import cached_property

from searchkit.constraints import TimestampMatcherBase
from hotsos.core.host_helpers import UptimeHelper, CLIHelper
from hotsos.core.log import log
from hotsos.core.search import (
    SearchDef,
    SequenceSearchDef,
    SearchConstraintSearchSince,
)
from hotsos.core.ycheck.engine.properties.common import (
    cached_yproperty_attr,
    YPropertyOverrideBase,
    YPropertyMappedOverrideBase,
    add_to_property_catalog,
)


class CommonTimestampMatcher(TimestampMatcherBase):
    """
    This class must support regex patterns to match any kind of timestamp that
    we would expect to find. When a plugin results in the search of file
    (typically log files) that contain timestamps it is necessary to ensure
    the patterns in this class support matching those timestamps in order for
    search constraints to work.

    TODO: timestamps typically use an RFC format so we should brand them
          as such here.
    """
    MONTH_MAP = {'jan': 1, 'feb': 2, 'mar': 3, 'apr': 4, 'may': 5,
                 'jun': 6, 'jul': 7, 'aug': 8, 'sep': 9, 'oct': 10,
                 'nov': 11, 'dec': 12}

    @cached_property
    def _current_year(self):
        return CLIHelper().date(format='+%Y')

    @property
    def year(self):
        """ Needed for kernlog which has no year group. """
        try:
            return self.result.group('year')
        except IndexError:
            pass

        return self._current_year

    @property
    def month(self):
        """ Needed for kernlog which has a string month. """
        try:
            return int(self.result.group('month'))
        except ValueError:
            pass

        _month = self.result.group('month').lower()
        try:
            return self.MONTH_MAP[_month[:3]]
        except KeyError:
            log.exception("could not establish month integer from '%s'",
                          _month)

    @property
    def patterns(self):
        """
        This needs to contain timestamp patterns for any/all types of file
        we want to analyse where SearchConstraintsSince is to be applied.
        """
        # should match plugins.openstack.openstack.OpenstackDateTimeMatcher
        openstack = (r'^(?P<year>\d{4})-(?P<month>\d{2})-(?P<day>\d{2})+\s+'
                     r'(?P<hours>\d{2}):(?P<minutes>\d{2}):(?P<seconds>\d+)')
        # since they are identical we wont add but leaving in case we want to.
        # edit later.
        # juju = openstack
        # should match plugins.storage.ceph.CephDateTimeMatcher
        ceph = (r'^(?P<year>\d{4})-(?P<month>\d{2})-(?P<day>\d{2})[\sT]'
                r'(?P<hours>\d{2}):(?P<minutes>\d{2}):(?P<seconds>\d+)')
        # since they are identical we wont add but leaving in case we want to
        # edit later.
        # openvswitch = ceph
        kernlog = (r'^(?P<month>\w{3,5})\s+(?P<day>\d{1,2})\s+'
                   r'(?P<hours>\d{2}):(?P<minutes>\d{2}):(?P<seconds>\d{2})')
        return [openstack, ceph, kernlog]


class YPropertySearchConstraints(YPropertyOverrideBase):

    @classmethod
    def _override_keys(cls):
        return ['constraints']

    @property
    def valid_attributes(self):
        return ['search-period-hours', 'search-result-age-hours',
                'min-hours-since-last-boot', 'min-results']

    @cached_yproperty_attr
    def search_period_hours(self):
        """
        If min is provided this is used to determine the period within which
        min applies. If period is unset, the period is infinite i.e. across all
        available data.

        Supported values:
          <int> hours

        """
        return int(self.content.get('search-period-hours', 0))

    @cached_yproperty_attr
    def search_result_age_hours(self):
        """
        Result must have occurred within this number of hours from the current
        time (for a sosreport this would be when it was created).
        """
        return int(self.content.get('search-result-age-hours', 0))

    @cached_yproperty_attr
    def min_hours_since_last_boot(self):
        """
        Search result must be at least this number of hours after the last
        boot time.
        """
        return int(self.content.get('min-hours-since-last-boot', 0))

    @cached_yproperty_attr
    def min_results(self):
        """
        Minimum search matches required for result to be True (default is 1)
        """
        return int(self.content.get('min-results', 1))

    @cached_yproperty_attr
    def filesearch_constraints_obj(self):
        """
        Create a search constraints object representing the paramaters in
        this property.
        """
        invalid = []
        for attr in self.content:
            if attr not in self.valid_attributes:
                invalid.append(attr)

        if invalid:
            raise Exception("Invalid search constraints attributes found: {}. "
                            "Valid options are: {}".
                            format(', '.join(invalid),
                                   ', '.join(self.valid_attributes)))

        has_result_hours = 'search-result-age-hours' in self.content
        has_boot_hours = 'min-hours-since-last-boot' in self.content
        if not any([has_result_hours, has_boot_hours]):
            return

        uptime_etime_hours = UptimeHelper().in_hours
        hours = self.search_result_age_hours
        min_hours_since_last_boot = self.min_hours_since_last_boot
        if not hours:
            hours = max(uptime_etime_hours - min_hours_since_last_boot, 0)
        elif min_hours_since_last_boot > 0:  # pylint: disable=W0143
            hours = min(hours,
                        max(uptime_etime_hours - min_hours_since_last_boot, 0))

        return SearchConstraintSearchSince(
                                         ts_matcher_cls=CommonTimestampMatcher,
                                         hours=hours)


class YPropertySearchOpt(YPropertyOverrideBase):

    @classmethod
    def _override_keys(cls):
        return ['expr', 'hint', 'passthrough-results']

    def __bool__(self):
        return bool(self.content)

    @property
    def expr(self):
        """ Can be string or list. """
        return self.content

    def __str__(self):
        # should use bool() for passthrough-results
        invalid = ['passthrough-results', 'expr']
        valid = [k for k in self._override_keys() if k not in invalid]
        if self._override_name not in valid:
            raise Exception("__str__ only valid for {} (not {})".
                            format(','.join(valid),
                                   self._override_name))

        return self.content


class YPropertySearchBase(YPropertyMappedOverrideBase):

    @classmethod
    def _override_mapped_member_types(cls):
        return [YPropertySearchOpt, YPropertySearchConstraints]

    @classmethod
    def get_datetime_from_result(cls, result):
        """
        This attempts to create a datetime object from a timestamp (usually
        from a log file) extracted from a search result. If it is not able
        to do so it will return None. The normal expectation is that two search
        result groups be available at index 1 and 2 but if only 1 is valid it
        will be used a fallback.
        """
        ts = result.get(1)
        if result.get(2):
            ts = "{} {}".format(ts, result.get(2))
        else:
            ts = "{} 00:00:00".format(ts)

        ts_matcher = CommonTimestampMatcher(ts)
        if ts_matcher.matched:
            return ts_matcher.strptime

        log.warning("failed to parse timestamp string '%s' (num_group=%s) - "
                    "returning None", ts, len(result))

    @classmethod
    def filter_by_period(cls, results, period_hours):
        """ Return the most recent period_hours worth of results. """
        if not period_hours:
            log.debug("period filter not specified - skipping")
            return results

        log.debug("applying search filter (period_hours=%s)", period_hours)

        _results = []
        for r in results:
            ts = cls.get_datetime_from_result(r)
            if ts:
                _results.append((ts, r))

        results = []
        last = None

        for r in sorted(_results, key=lambda i: i[0], reverse=True):
            if last is None:
                last = r[0]
            elif r[0] < last - timedelta(hours=period_hours):
                break

            results.append(r)

        log.debug("%s results remain after applying filter", len(results))
        return [r[1] for r in results]

    def apply_extra_constraints(self, results):
        """
        Apply further constraints filtering to search results.

        These are constraints supported by the constraints property that
        are not/cannot be applied by the search engine itself using
        SearchConstraintSearchSince.
        """
        if not self.constraints:
            log.debug("no extra search constraints to apply")
            return results

        log.debug("applying extra search constraints")
        if results:
            period_hours = self.constraints.search_period_hours
            results = self.filter_by_period(results, period_hours)

        count = len(results)
        if count < self.constraints.min_results:
            log.debug("search does not have enough matches (%s) to "
                      "satisfy min of %s", count, self.constraints.min_results)
            return []

        log.debug("applying extra search constraints reduced results from %s "
                  "to %s", count, len(results))
        return results

    @property
    def unique_search_tag(self):
        return self._override_path

    @property
    def passthrough_results_opt(self):
        if self.passthrough_results is not None:
            return bool(self.passthrough_results)

        return False

    def _resolve_exprs(self, patterns):
        """
        Resolve any expressions provided as variable to their value. Non-vars
        are returned as-is.
        """
        _patterns = []
        if type(patterns) == list:
            for p in patterns:
                _patterns.append(self.resolve_var(p))
        else:
            _patterns = self.resolve_var(patterns)

        return _patterns

    @property
    def search_pattern(self):
        if self.expr:
            return self._resolve_exprs(self.expr.expr)

        # can be supplied as a single string or list of strings
        patterns = self._resolve_exprs(list(self.content.keys()))
        if len(patterns) == 0:
            raise Exception("no search pattern (expr) defined")

        for pattern in patterns:
            if not re.search(r"[^\\]\(", pattern):
                log.info("pattern '%s' does not contain a subgroup. this "
                         "is inefficient and can result in unnecessary "
                         "memory consumption", pattern)

        if len(patterns) == 1:
            return patterns[0]

        return patterns

    @property
    def is_sequence_search(self):
        seq_keys = YPropertySequencePart._override_keys()
        return any(getattr(self, key) for key in seq_keys)

    @property
    def simple_search(self):
        if (self.is_sequence_search or not self.search_pattern or
                self.passthrough_results_opt):
            return

        sdef = self.cache.simple_search
        if sdef:
            return sdef

        pattern = self.search_pattern
        hint = None
        if self.hint:
            hint = str(self.hint)

        constraints = None
        if self.constraints and self.constraints.filesearch_constraints_obj:
            constraints = [self.constraints.filesearch_constraints_obj]

        sdef = SearchDef(pattern, tag=self.unique_search_tag, hint=hint,
                         constraints=constraints)

        self.cache.set('simple_search', sdef)
        return sdef

    @property
    def sequence_search(self):
        if not self.is_sequence_search or self.passthrough_results_opt:
            return

        sdef = self.cache.sequence_search
        if sdef:
            return sdef

        seq_start = self.start
        seq_body = self.body
        seq_end = self.end

        if (seq_body or (seq_end and not self.passthrough_results_opt)):
            sd_start = SearchDef(seq_start.search_pattern)

            sd_end = None
            # explicit end is optional for sequence definition
            if seq_end:
                sd_end = SearchDef(seq_end.search_pattern)

            sd_body = None
            if seq_body:
                sd_body = SearchDef(seq_body.search_pattern)

            # NOTE: we don't use hints here
            tag = self.unique_search_tag
            sdef = SequenceSearchDef(start=sd_start, body=sd_body,
                                     end=sd_end, tag=tag)
            self.cache.set('sequence_search', sdef)
            return sdef

        log.warning("invalid sequence definition passthrough=%s "
                    "start=%s, body=%s, end=%s",
                    self.passthrough_results_opt, seq_start, seq_body, seq_end)

    @property
    def sequence_passthrough_search(self):
        if not self.is_sequence_search or not self.passthrough_results_opt:
            return

        sdef = self.cache.sequence_passthrough_search
        if sdef:
            return sdef

        seq_start = self.start
        seq_end = self.end

        if self.passthrough_results_opt and all([seq_start, seq_end]):
            # start and end required for core.analytics.LogEventStats
            start_tag = "{}-start".format(self.unique_search_tag)
            end_tag = "{}-end".format(self.unique_search_tag)
            sdefs = [SearchDef(str(seq_start.search_pattern), tag=start_tag),
                     SearchDef(str(seq_end.search_pattern), tag=end_tag)]
            self.cache.set('sequence_passthrough_search', sdefs)
            return sdefs

    def load_searcher(self, searchobj, search_path, allow_constraints=True):
        """ Load search definitions into the given searcher object. """
        sdef = self.simple_search
        if sdef:
            log.debug("loading simple search")
            searchobj.add(sdef, search_path,
                          allow_global_constraints=allow_constraints)
            return

        sdef = self.sequence_search
        if sdef:
            log.debug("loading sequence search")
            searchobj.add(sdef, search_path,
                          allow_global_constraints=allow_constraints)
            return

        sdef = self.sequence_passthrough_search
        if sdef:
            log.debug("loading sequence passthrough searches")
            for _sdef in sdef:
                searchobj.add(_sdef, search_path,
                              allow_global_constraints=allow_constraints)


class YPropertySequencePart(YPropertySearchBase):

    @classmethod
    def _override_keys(cls):
        return ['start', 'body', 'end']

    def __str__(self):
        for stack in self._stack.current.content.values():
            return stack.current.content


@add_to_property_catalog
class YPropertySearch(YPropertySearchBase):

    @classmethod
    def _override_keys(cls):
        return ['search']

    @classmethod
    def _override_mapped_member_types(cls):
        members = super()._override_mapped_member_types()
        return members + [YPropertySequencePart]



from hotsos.core.plugins.vault import VaultChecksBase
from hotsos.core.plugintools import summary_entry_offset as idx


class VaultSummary(VaultChecksBase):

    @idx(0)
    def __summary_services(self):
        if self.systemd.services:
            return self.systemd.summary
        if self.pebble.services:
            return self.pebble.summary

    @idx(1)
    def __summary_snaps(self):
        if self.snaps.core:
            return self.snaps.all_formatted
